// PackPacker.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "zip.h"
#include "ioapi.h"
#include "iowin32.h"
#include "j2linfo.h"

#define WRITE_BUFFER_SIZE	4096
#define ERR(format, ...)	fwprintf(stderr, format, __VA_ARGS__)

static BOOL directoryExists(const wchar_t* dirName)
{
	DWORD attribs = GetFileAttributes(dirName);
	if (attribs == INVALID_FILE_ATTRIBUTES) {
		return FALSE;
	}
	return (attribs & FILE_ATTRIBUTE_DIRECTORY);
}

static BOOL fileExists(const wchar_t* fileName)
{
	DWORD attribs = GetFileAttributes(fileName);
	if (attribs == INVALID_FILE_ATTRIBUTES) {
		return FALSE;
	}
	return (!(attribs & FILE_ATTRIBUTE_DIRECTORY));
}

static uLong filetime(f, tmzip, dt)
	wchar_t *f;		/* name of file to get info on */
	tm_zip *tmzip;	/* return value: access, modific. and creation times */
	uLong *dt;		/* dostime */
{
	FILETIME ftLocal;
	HANDLE hFind;
	WIN32_FIND_DATA  ff32;

	hFind = FindFirstFile(f, &ff32);
	if (hFind != INVALID_HANDLE_VALUE)
	{
		FileTimeToLocalFileTime(&(ff32.ftLastWriteTime), &ftLocal);
		FileTimeToDosDateTime(&ftLocal, ((LPWORD)dt)+1, ((LPWORD)dt)+0);
		FindClose(hFind);
		return 1;
	}
	return 0;
}

static int getFilename(const wchar_t *path, char* filename, int count)
{
	int ret;
	wchar_t drive[_MAX_DRIVE],
		dir[_MAX_DIR],
		fname[_MAX_FNAME],
		ext[_MAX_EXT],
		completeFn[_MAX_FNAME + _MAX_EXT];

	ret = _wsplitpath_s(
		path, drive, _MAX_DRIVE, dir, _MAX_DIR, fname, _MAX_FNAME,
		ext, _MAX_EXT);
	if (ret != 0)
		return 1;

	if (wcscpy_s(completeFn, _MAX_FNAME + _MAX_EXT, fname) != 0)
		return 1;
	if (wcscat_s(completeFn, _MAX_FNAME + _MAX_EXT, ext) != 0)
		return 1;

	wcstombs(filename, completeFn, count);

	return 0;
}

int createZip(wchar_t *zipPath, listP filePaths, BOOL overwrite)
{
	zlib_filefunc_def	ffunc;
	int					i,
						size_buf = WRITE_BUFFER_SIZE;
	zipFile				zf;
	char				buf[WRITE_BUFFER_SIZE];

	fill_win32_filefunc(&ffunc);

	if (fileExists(zipPath) && !overwrite) {
		ERR(L"Zip file path %s already exists.\n", zipPath);
		return 3;
	}
	zf = zipOpen2(zipPath, APPEND_STATUS_CREATE, NULL, &ffunc);
	if (zf == NULL) {
		ERR(L"Could not open zip file for writing.\n");
		return 4;
	}

	//add the several files
	for (i = 0; i < listSize(filePaths); i++) {
		wchar_t			*fileInZip = listElementAt(filePaths, i);
		FILE			*fin;
		char			filenameinzip[1024];
		zip_fileinfo	zi;
		unsigned long	crcFile=0;
		int				err = ZIP_OK;

		if (getFilename(fileInZip, filenameinzip, 1024) != 0) {
			ERR(L"Could not extract filename from %s\n", fileInZip);
			err = ZIP_ERRNO;
		}

		memset(&zi, '\0', sizeof zi);
		if (err == ZIP_OK) {
			filetime(fileInZip, &zi.tmz_date, &zi.dosDate); //extract time

			err = zipOpenNewFileInZip(zf, filenameinzip, &zi,
				NULL, 0, NULL, 0, NULL /* comment */,
				Z_DEFLATED, 9 /* compression level */);

			if (err != ZIP_OK) {
				ERR(L"Error in opening %s in zipfile\n", filenameinzip);
			} else {
				fin = _wfopen(fileInZip, L"rb");
				if (fin == NULL) {
					ERR(L"Error in opening %s for reading\n", filenameinzip);
					err = ZIP_ERRNO;
				}
			}
		}

		if (err == ZIP_OK) { //if the original and in zip were correcyly opened
			//write the data
			int size_read;
			do {
				err = ZIP_OK;
				size_read = (int) fread(buf, 1, size_buf, fin);
				if (size_read < size_buf)
					if (feof(fin) == 0) {
						ERR(L"Error in reading %s.\n", filenameinzip);
						err = ZIP_ERRNO;
					}

					if (size_read > 0) {
						err = zipWriteInFileInZip(zf, buf, size_read);
						if (err < 0) {
							ERR(L"Error in writing %s in the zipfile.\n",
								filenameinzip);
							err = ZIP_ERRNO;
						}
					}
			} while ((err == ZIP_OK) && (size_read > 0)); 
		}

		/*
		 * We may have stopped:
		 * - When failing to open the file in zip
		 * - When failing to open the original file
		 * - Halfway through writing the file in zip
		 */

        if (fin != NULL)
            fclose(fin);

		if (err < 0) {
			err = ZIP_ERRNO;
		} else { //only close file in zio if there were no errors
			err = zipCloseFileInZip(zf);
			if (err != ZIP_OK)
				ERR(L"Error in closing %s in the zipfile.\n", filenameinzip);
		}
	} //end of for

	{ //close zip file
		int errclose = zipClose(zf,NULL);
		if (errclose != ZIP_OK) {
			ERR(L"Error in closing %s.\n", zipPath);
			return 1;
		}
    }
	
	return 0;
}

//strings better be null-determinated, or else...
static BOOL listContainsStringI(listP list, const wchar_t *string)
{
	int i;
	for (i = 0; i < listSize(list); i++) {
		wchar_t *elem = listElementAt(list, i);
		if (elem != NULL && _wcsicmp(string, elem) == 0)
			return TRUE;
	}

	return FALSE;
}

//caller is responsible to make sure name is strlen(name)+1+4 long
//extensions are assumed to be .xxx or smaller
static void ensureExtension(wchar_t *name,
							int size,
							wchar_t *extension, //with dot
							BOOL lax) //do not add ext if we already have one
{
	wchar_t *f, *g;
	BOOL endsWithGivenExtension;

	if (lax && wcschr(name, L'.') != NULL)
		return; //do nothing if lax and we already have an extension

	if (wcslen(extension) > wcslen(name))
		endsWithGivenExtension = FALSE;

	f = name + wcslen(name); //f now points to a L'\0' character
	g = extension + wcslen(extension); //idem for g
	endsWithGivenExtension = TRUE;
	while (g >= extension) {
		if (*f != *g) {
			endsWithGivenExtension = FALSE;
			break;
		}
		
		f--;
		g--;
	}

	if (!endsWithGivenExtension)
		wcscat_s(name, size, extension);
}

#define ADD_IF_NEW_EXISTS_IGNORE				0
#define	ADD_IF_NEW_EXISTS_NEW_EXISTS			1
#define ADD_IF_NEW_EXISTS_NEW_DOES_NOT_EXIST	2

static int copyToNewString(wchar_t **dest, const wchar_t *src)
{
	*dest = calloc(wcslen(src) + 1, sizeof **dest);
	wcscpy(*dest, src);

	return 0;
}

static int copyMbToWcNewString(wchar_t **dest, const char *src)
{
	size_t len = _mbslen(src);
	*dest = calloc(len + 1, sizeof *dest);
	if (mbstowcs(*dest, src, len) == -1) {
		free(*dest);
		return 1; //failure
	} else {
		return 0; //success
	}
}

static int addIfNewAndExists(const wchar_t *file,
							 listP path,
							 listP skipList,
							 listP dependencies,
							 listP alreadySeen,
							 const wchar_t **filepath) //OUT
{
	wchar_t *filecopy;
	BOOL	fileFound = FALSE;
	int		i;
	size_t	fileLen;

	if (filepath != NULL)
		*filepath = NULL;
	
	fileLen = wcslen(file);

	if (fileLen == 0 || listContainsStringI(alreadySeen, file))
		return ADD_IF_NEW_EXISTS_IGNORE;
	
	copyToNewString(&filecopy, file);
	listAdd(alreadySeen, filecopy);

	for (i = 0; i < listSize(path) && !fileFound; i++) {
		wchar_t *curDir,
				*dirPlusFile;
		size_t	curDirLen;

		curDir = listElementAt(path, i);
		curDirLen = wcslen(curDir);
		dirPlusFile = calloc(fileLen + curDirLen + 2, sizeof *dirPlusFile);
		wcscpy_s(dirPlusFile, fileLen + curDirLen + 2, curDir);
		wcscat_s(dirPlusFile, fileLen + curDirLen + 2, L"\\");
		wcscat_s(dirPlusFile, fileLen + curDirLen + 2, file);

		if (fileExists(dirPlusFile)) {
			if (!listContainsStringI(skipList, filecopy)) { //is not to be skipped
				listAdd(dependencies, dirPlusFile);
			}
			if (filepath != NULL) {
				*filepath = dirPlusFile;
			}
			fileFound = TRUE;
		} else {
			free(dirPlusFile);
		}
	}

	if (!fileFound)
		return ADD_IF_NEW_EXISTS_NEW_DOES_NOT_EXIST;

	return ADD_IF_NEW_EXISTS_NEW_EXISTS;
}
static int addIfNewAndExistsChar(const char *file,
								 listP path,
								 listP skipList,
								 listP dependencies,
								 listP alreadySeen,
								 wchar_t **filepath)
{
	//a bit big, since the strings it will receive will be no longer than 36
	wchar_t wcfile[_MAX_PATH];

	mbstowcs(wcfile, file, _MAX_PATH);
	wcfile[_MAX_PATH - 1] = L'\0'; //protect agains files too big
	return addIfNewAndExists(wcfile, path, skipList, dependencies,
		alreadySeen, filepath);
}


static void dependSearch(listP initialFiles,
						 listP skipList,
						 listP path,
						 listP *allDeps,
						 listP *validDeps)
{
	int			i;
	listP		dependencies = listNew(20), //valid dependencies (full path)
				alreadySeen = listNew(20); //valid & missing depend. (filename)
	stackP		levelsToSearch = stackNew(); //aloc before push, free after pop	

	//Fill the stack with initial files
	for (i = 0; i < listSize(initialFiles); i++) {
		wchar_t *file;
		if (wcslen(listElementAt(initialFiles, i)) > J2F_MAX_LEVEL_NAME_SIZE) {
			ERR(L"Jazz2 files cannot have a name bigger than %i.\n",
				J2F_MAX_LEVEL_NAME_SIZE);
		}
		copyToNewString(&file, listElementAt(initialFiles, i));
		stackPush(levelsToSearch, file);
	}

	while (!stackIsEmpty(levelsToSearch)) {
		wchar_t	*tmpstr;
		int		res;
		//+ 4 + 1 to append .j2l if needed
#define J2_FILE_AND_EXTENSION (J2F_MAX_LEVEL_NAME_SIZE + 4 + 1)
		wchar_t		lvlname[J2_FILE_AND_EXTENSION],
					buffer[J2_FILE_AND_EXTENSION],
					*filepath;
		LevelInfo	levelInfo;

		tmpstr = stackPop(levelsToSearch);
		wcscpy_s(lvlname, J2_FILE_AND_EXTENSION, tmpstr);
		free(tmpstr);
		ensureExtension(lvlname, J2_FILE_AND_EXTENSION, L".j2l",
			FALSE);

		res = addIfNewAndExists(lvlname, path, skipList, dependencies,
			alreadySeen, &filepath);
		if (res == ADD_IF_NEW_EXISTS_IGNORE)
			continue; //skip if already processed

		if (res == ADD_IF_NEW_EXISTS_NEW_DOES_NOT_EXIST) {
			ERR(L"Level dependency does not exist: %s.\n", lvlname);
			continue;
		}
		
		if (readLevelInformation(filepath, &levelInfo) != J2F_SUCCESS) {
			ERR(L"Failing to read information from level %s.\n", lvlname);
			continue;
		}
		
		if (mbstowcs(buffer, levelInfo.tileSet, J2_FILE_AND_EXTENSION)) {
			ensureExtension(buffer, J2_FILE_AND_EXTENSION, L".j2t", FALSE);
			if (addIfNewAndExists(buffer, path, skipList, dependencies,
				alreadySeen, &filepath) == ADD_IF_NEW_EXISTS_NEW_DOES_NOT_EXIST)
				ERR(L"Tileset %s for level %s is missing.\n",
				buffer, lvlname);
		} else {
			ERR(L"Failure to convert tileset name to wide char string.");
		}

		if (mbstowcs(buffer, levelInfo.musicFile, J2_FILE_AND_EXTENSION)) {
			ensureExtension(buffer, J2_FILE_AND_EXTENSION, L".j2b", TRUE);
			if (addIfNewAndExists(buffer, path, skipList, dependencies,
				alreadySeen, &filepath) == ADD_IF_NEW_EXISTS_NEW_DOES_NOT_EXIST)
				ERR(L"Music file %s for level %s is missing.\n",
				buffer, lvlname);
		} else {
			ERR(L"Failure to convert music file name to wide char string.");
		}

		if (strlen(levelInfo.nextLevel) > 0) {
			if (copyMbToWcNewString(&tmpstr, levelInfo.nextLevel) == 0)
				stackPush(levelsToSearch, tmpstr);
			else
				ERR(L"Failure to convert next level to wide char string.");
		}

		if (strlen(levelInfo.secretLevel) > 0) {
			if (copyMbToWcNewString(&tmpstr, levelInfo.secretLevel) == 0)
				stackPush(levelsToSearch, tmpstr);
			else
				ERR(L"Failure to convert secret level to wide char string.");
		}

		if (strlen(levelInfo.bonusLevel) > 0) {
			if (copyMbToWcNewString(&tmpstr, levelInfo.bonusLevel) == 0)
			stackPush(levelsToSearch, tmpstr);
			else
				ERR(L"Failure to convert bonus level to wide char string.");
		}
	}

	//out arguments
	*allDeps	= alreadySeen;
	*validDeps	= dependencies;
}

static void dependSearchDestroy(listP allDeps, listP validDeps)
{
	int i;
	for (i = 0; i < listSize(allDeps); i++) {
		free(listElementAt(allDeps, i));
	}
	for (i = 0; i < listSize(validDeps); i++) {
		free(listElementAt(validDeps, i));
	}
	listDestroy(allDeps);
	listDestroy(validDeps);
}

static void showSyntax(void)
{
	ERR(L"PackPacker 2009-08-01 by Cataphract\n");
	ERR(L"Syntax:	/Z <zip> [/O] [/S <file>] [/D <dir> [/D <dir> ...]]\n"
		L"<level> [level [level ...]]\n\n");
	ERR(L"	/Z	The name of the zip file.\n");
	ERR(L"	/O	Overwrites the zip file if it already exists.\n");
	ERR(L"	/S	Excludes from zip files given in a text file.\n");
	ERR(L"	/D	Specifies additional directories to look for files.\n");
	ERR(L"Further arguments indicate the levels with which to seed the "
		L"dependency\nsearch. The .j2l extension is not required.\n");
	ERR(L"If no path is specified, the working directory will be used.");
}

BOOL parseArguments(int argc,					//IN
					const wchar_t *argv[],		//IN
					wchar_t const**zipPath,		//OUT, max size _MAX_PATH
					const wchar_t **skipList,	//OUT, max size _MAX_PATH
					listP seedLevels,			//OUT, initial size 0
					listP paths,				//OUT, initial size 0
					BOOL *overwrite)			//OUT
{
	BOOL expectingParamVal = FALSE;
	const wchar_t *currentOption;
	int i;

	*overwrite	= FALSE;
	*zipPath	= NULL;
	*skipList	= NULL;
	
	for (i = 1; i < argc; i++) {
		//option specified
		if (wcslen(argv[i]) > 1 && argv[i][0] == L'/') {
			if (expectingParamVal) {
				ERR(L"Expecting parameter for option %s, "
					L"but new option %s was given.\n", currentOption, argv[i]);
				return FALSE;
			}
			if (_wcsicmp(L"/O", argv[i]) == 0) {
				*overwrite = TRUE;
			} else if (_wcsicmp(L"/S", argv[i]) == 0 ||
				_wcsicmp(L"/D", argv[i]) == 0 || _wcsicmp(L"/Z", argv[i]) == 0) {
				currentOption = argv[i];
				expectingParamVal = TRUE;
			} else {
				ERR(L"Unrecognized option: %s.\n", argv[i]);
				return FALSE;
			}
		} else {
			if (!expectingParamVal) {
				listAdd(seedLevels, (wchar_t *) argv[i]);
			} else if (_wcsicmp(L"/S", currentOption) == 0) {
				if (wcslen(argv[i]) > (_MAX_PATH -1)) {
					ERR(L"Skip list argument is too long: %s.\n", argv[i]);
					return FALSE;
				}
				*skipList = argv[i];
				expectingParamVal = FALSE;
			} else if (_wcsicmp(L"/D", currentOption) == 0) {
				listAdd(paths, (wchar_t *) argv[i]);
				expectingParamVal = FALSE;
			} else if (_wcsicmp(L"/Z", currentOption) == 0) {
				if (wcslen(argv[i]) > (_MAX_PATH -1)) {
					ERR(L"Zip file argument is too long: %s.\n", argv[i]);
					return FALSE;
				}
				*zipPath = argv[i];
				expectingParamVal = FALSE;
			}
		}
	} //end for
	if (listSize(paths) == 0) {
		//if no directory was given, the current one is searched
		listAdd(paths, L".");
	}
	if (listSize(seedLevels) == 0) {
		ERR(L"At least one seed level must be given.\n");
		return FALSE;
	}
	if (*zipPath == NULL) {
		ERR(L"The /Z option was not specified.\n");
		return FALSE;
	}

	return TRUE;
}

static int validatePaths(const wchar_t *zipPath,
						 const wchar_t *skipList,
						 listP seedLevels,
						 listP paths)
{
	int i;

	if (directoryExists(zipPath)) {
		ERR(L"There is already a directory with the name you "
			L"specified for the zip file: %s.\n", *zipPath);
		return FALSE;
	}
	if (skipList != NULL && !fileExists(skipList)) {
		ERR(L"The file for the skip list does not exist: %s.\n", skipList);
		return FALSE;
	}
	for (i = 0; i < listSize(seedLevels); i++) {
		wchar_t *lvlname = listElementAt(seedLevels, i);
		while (*lvlname != L'\0') {
			if (*lvlname == L'\\' || *lvlname == L'/' || *lvlname == L':') {
				ERR(L"Level names must not include a path component. "
					L"Use the /D option instead. %s was unaccepted.\n",
					listElementAt(seedLevels, i));
				return FALSE;
			}
			lvlname++;
		}
	}
	for (i = 0; i < listSize(paths); i++) {
		if (!directoryExists(listElementAt(paths, i))) {
			ERR(L"There is no such directory: %s.\n", listElementAt(paths, i));
			return FALSE;
		}
	}

	return TRUE;
}

static int readSkipListFile(const wchar_t *skipListFile,
							listP skipList)
{
	FILE	*f;
	wchar_t	buffer[1024];

	if ((f = _wfopen(skipListFile, L"rS")) == NULL) {
		ERR(L"Could not open skip file %s for reading.\n", skipListFile);
		return 1;
	}
	while (TRUE) {
		if (fgetws(buffer, 1024, f) != NULL) {
			wchar_t	*fileToAdd	= malloc(1024 * sizeof fileToAdd),
					*sptr		= buffer,
					*dptr		= fileToAdd;
			if (buffer[0] == L'\n') { //blank line
				free(fileToAdd);
				continue;
			}
			while (*sptr != L'\0' && *sptr != L'*' && *sptr != L'\n') {
				*dptr = *sptr;
				*sptr++;
				*dptr++;
			}
			*dptr = L'\0';
			listAdd(skipList, fileToAdd);
		} else if (!feof(f)) {
			ERR(L"Error reading skip file %s.\n", skipListFile);
			return 1;
		} else { //EOF
			break;
		}
	}

	return 0;
}

static void skipListDestroy(listP skipList)
{
	int i;
	for (i = 0; i < listSize(skipList); i++) {
		free(listElementAt(skipList, i));
	}
}

int _tmain(int argc, wchar_t *argv[])
{
	int		zipRet,
			ret = -1;
	wchar_t	*zipPath,
			*skipListFile;
	listP	allDeps		= NULL,
			validDeps	= NULL,
			paths		= listNew(5),
			seedLevels	= listNew(5),
			skipList	= listNew(40);
	BOOL	overwrite;

	if (!parseArguments(argc, argv, &zipPath, &skipListFile, seedLevels, paths,
		&overwrite)) {
			ERR(L"Invalid arguments.\n\n");
			showSyntax();
			goto err;
	}
	if (!validatePaths(zipPath, skipListFile, seedLevels, paths))
		goto err;

	if (skipListFile != NULL) {
		if (readSkipListFile(skipListFile, skipList) != 0) {
			ERR(L"Error reading skip list file %s.\n", skipListFile);
			goto err;
		}
	}

	dependSearch(seedLevels, skipList, paths, &allDeps, &validDeps);

	zipRet = createZip(zipPath, validDeps, overwrite);

	if (zipRet == 3) {
		ERR(L"No action taken, as the zip file already existed.\n");
	} else if (zipRet != 0) {
		ERR(L"There was an error during zip file creation.\n");
	}

	ret = 0;
err:
	listDestroy(paths);
	listDestroy(seedLevels);
	skipListDestroy(skipList);
	listDestroy(skipList);
	if (allDeps != NULL)
		dependSearchDestroy(allDeps, validDeps);

	return (ret == -1) ? 1 : ret;
}
