#include "stdafx.h"

/*
 * Linked-list implementation of stacks.
 * Constant time for everything but destroyStack().
 */

typedef struct stack_node *stackNodeP;
struct stack_node {
	void		*elem;
	stackNodeP	previous;
};

struct stack_header {
	long		numElements;
	stackNodeP	last;
};

stackP stackNew(void)
{
	return calloc(1, sizeof(struct stack_header));
}

static void destroyNodes(stackNodeP node)
{
	stackNodeP prevNode;

	if (node == NULL)
		return;
	prevNode = node->previous;
	free(node);
	destroyNodes(prevNode); //tail recursion
}

void stackDestroy(stackP stack)
{
	stackNodeP lastNode = stack->last;
	free(stack);
	destroyNodes(lastNode);
}

void *stackPop(stackP stack)
{
	stackNodeP	retNode;
	void*		retVal;

	if (stack->numElements == 0)
		return NULL;
	retNode = stack->last;
	retVal = retNode->elem;
	
	stack->last = retNode->previous;
	free(retNode);
	stack->numElements--;
	
	return retVal;
}

void stackPush(stackP stack, void *elem)
{
	stackNodeP newNode = calloc(1, sizeof *newNode);
	if (newNode == NULL)
		return; //fail silently...
	stack->numElements++;
	newNode->previous = stack->last;
	newNode->elem = elem;
	stack->last = newNode;
}

BOOL stackIsEmpty(stackP stack)
{
	return (stack->numElements == 0);
}

int	stackSize(stackP stack)
{
	return stack->numElements;
}

/*
 * Array list implementation
 */

struct list_struct {
	int capacity;
	int size;
	void **elems;
};

static BOOL adjustCapacity(listP list, int cap)
{
	if (cap < list->size)
		return FALSE;
	list->elems = _recalloc(list->elems, cap, sizeof *list->elems);
	if (list->elems != NULL) {
		list->capacity = cap;
		return TRUE;
	} else {
		return FALSE;
	}
}

listP listNew(int init_cap)
{
	listP list;
	if (init_cap < 1)
		return NULL;

	list = calloc(1, sizeof *list);
	if (list == NULL)
		return NULL;

	list->size = 0;
	if (adjustCapacity(list, init_cap))
		return list;
	else {
		free(list);
		return NULL;
	}

	return list;
}

void listDestroy(listP list)
{
	free(list->elems);
	free(list);
}

void listAddElements(listP list, void *elems[], int size)
{
	int i;
	if (list->size + size > list->capacity)
		adjustCapacity(list, (int) ceil(((double)(list->size + size)) * 1.5));
	for (i = 0; i < size; i++) {
		list->elems[list->size + i] = elems[i];
	}
	list->size += size;
}

void listAdd(listP list, void *elem)
{
	if (list->size == list->capacity)
		adjustCapacity(list, (int) ceil(((double)(list->size + 1)) * 1.5));
	list->elems[list->size] = elem;
	list->size++;
}

void *listElementAt(listP list, int index)
{
	if (index < 0 || index >= list->size)
		return NULL;
	return list->elems[index];
}

int listSize(listP list)
{
	return list->size;
}
